'use strict';

const postcss = require('postcss');
const parser = require('postcss-selector-parser');
const parse5 = require('parse5');
const shared = require('@tailwindcss-mangle/shared');
const babel = require('@babel/core');
const helperPluginUtils = require('@babel/helper-plugin-utils');
const MagicString = require('magic-string');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

const postcss__default = /*#__PURE__*/_interopDefaultCompat(postcss);
const parser__default = /*#__PURE__*/_interopDefaultCompat(parser);
const babel__default = /*#__PURE__*/_interopDefaultCompat(babel);
const MagicString__default = /*#__PURE__*/_interopDefaultCompat(MagicString);

function isObject(value) {
  return value !== null && typeof value === "object";
}
function _defu(baseObject, defaults, namespace = ".", merger) {
  if (!isObject(defaults)) {
    return _defu(baseObject, {}, namespace, merger);
  }
  const object = Object.assign({}, defaults);
  for (const key in baseObject) {
    if (key === "__proto__" || key === "constructor") {
      continue;
    }
    const value = baseObject[key];
    if (value === null || value === void 0) {
      continue;
    }
    if (merger && merger(object, key, value, namespace)) {
      continue;
    }
    if (Array.isArray(value) && Array.isArray(object[key])) {
      object[key] = [...value, ...object[key]];
    } else if (isObject(value) && isObject(object[key])) {
      object[key] = _defu(
        value,
        object[key],
        (namespace ? `${namespace}.` : "") + key.toString(),
        merger
      );
    } else {
      object[key] = value;
    }
  }
  return object;
}
function createDefu(merger) {
  return (...arguments_) => (
    // eslint-disable-next-line unicorn/no-array-reduce
    arguments_.reduce((p, c) => _defu(p, c, "", merger), {})
  );
}
const defu = createDefu();

const postcssPlugin = "postcss-mangle-tailwindcss-plugin";
function isVueScoped(s) {
  if (s.parent) {
    const index = s.parent.nodes.indexOf(s);
    if (index > -1) {
      const nextNode = s.parent.nodes[index + 1];
      if (nextNode && nextNode.type === "attribute" && nextNode.attribute.includes("data-v-")) {
        return true;
      }
    }
  }
  return false;
}
const transformSelectorPostcssPlugin = function(options) {
  const { ignoreVueScoped, replaceMap } = defu(options, {
    ignoreVueScoped: true
  });
  return {
    postcssPlugin,
    async Rule(rule) {
      await parser__default((selectors) => {
        selectors.walkClasses((s) => {
          if (s.value && replaceMap && replaceMap.has(s.value)) {
            if (ignoreVueScoped && isVueScoped(s)) {
              return;
            }
            const v = replaceMap.get(s.value);
            if (v) {
              s.value = v;
            }
          }
        });
      }).transform(rule, {
        lossless: false,
        updateSelector: true
      });
    }
  };
};
transformSelectorPostcssPlugin.postcss = true;

function cssHandler(rawSource, options) {
  const acceptedPlugins = [transformSelectorPostcssPlugin(options)];
  const { file } = options;
  return postcss__default(acceptedPlugins).process(rawSource, {
    from: file,
    to: file
  });
}

({
    HTML: parse5.html.NS.HTML,
    XML: parse5.html.NS.XML,
    MATHML: parse5.html.NS.MATHML,
    SVG: parse5.html.NS.SVG,
    XLINK: parse5.html.NS.XLINK,
    XMLNS: parse5.html.NS.XMLNS
});

/**
 * Determines if a given node is a document or not
 * @param {Node} node Node to test
 * @return {boolean}
 */
function isDocument(node) {
    return node.nodeName === '#document';
}
/**
 * Determines if a given node is a document fragment or not
 * @param {Node} node Node to test
 * @return {boolean}
 */
function isDocumentFragment(node) {
    return node.nodeName === '#document-fragment';
}
/**
 * Determines if a given node is a template node or not
 * @param {Node} node Node to test
 * @return {boolean}
 */
function isTemplateNode(node) {
    return node.nodeName === 'template';
}
const isElementNode = parse5.defaultTreeAdapter.isElementNode;
const isCommentNode = parse5.defaultTreeAdapter.isCommentNode;
const isDocumentTypeNode = parse5.defaultTreeAdapter.isDocumentTypeNode;
const isTextNode = parse5.defaultTreeAdapter.isTextNode;
/**
 * Determines if a given node is a parent or not
 * @param {Node} node Node to test
 * @return {boolean}
 */
function isParentNode(node) {
    return (isDocument(node) ||
        isDocumentFragment(node) ||
        isElementNode(node) ||
        isTemplateNode(node));
}

parse5.defaultTreeAdapter.appendChild;

/**
 * Traverses the tree of a given node
 * @param {Node} node Node to traverse
 * @param {Visitor} visitor Visitor to apply
 * @param {ParentNode=} parent Parent node of the current node
 * @return {void}
 */
function traverse(node, visitor, parent) {
    const shouldVisitChildren = typeof visitor['pre:node'] !== 'function' ||
        visitor['pre:node'](node, parent) !== false;
    if (shouldVisitChildren && isParentNode(node)) {
        for (const child of node.childNodes) {
            traverse(child, visitor, node);
        }
    }
    if (typeof visitor.node === 'function') {
        visitor.node(node, parent);
    }
    if (typeof visitor.document === 'function' && isDocument(node)) {
        visitor.document(node);
    }
    if (typeof visitor.documentFragment === 'function' &&
        isDocumentFragment(node)) {
        visitor.documentFragment(node, parent);
    }
    if (typeof visitor.element === 'function' && isElementNode(node)) {
        visitor.element(node, parent);
    }
    if (typeof visitor.template === 'function' && isTemplateNode(node)) {
        visitor.template(node, parent);
    }
    if (typeof visitor.comment === 'function' && isCommentNode(node)) {
        visitor.comment(node, parent);
    }
    if (typeof visitor.text === 'function' && isTextNode(node)) {
        visitor.text(node, parent);
    }
    if (typeof visitor.documentType === 'function' && isDocumentTypeNode(node)) {
        visitor.documentType(node, parent);
    }
}

function htmlHandler(rawSource, options) {
  const { replaceMap, classGenerator } = options;
  const fragment = parse5.parse(rawSource);
  traverse(fragment, {
    element(node) {
      const attribute = node.attrs.find((x) => x.name === "class");
      if (attribute) {
        const array = shared.splitCode(attribute.value, {
          splitQuote: false
        });
        for (const v of array) {
          if (replaceMap.has(v)) {
            attribute.value = attribute.value.replace(shared.makeRegex(v), classGenerator.generateClassName(v).name);
          }
        }
      }
    }
  });
  return parse5.serialize(fragment);
}

const isProd = () => process.env.NODE_ENV === "production";

// >>> INTERFACES <<<
// >>> HELPERS <<<
var castComparer = function (comparer) { return function (a, b, order) { return comparer(a, b, order) * order; }; };
var throwInvalidConfigErrorIfTrue = function (condition, context) {
    if (condition)
        throw Error("Invalid sort config: " + context);
};
var unpackObjectSorter = function (sortByObj) {
    var _a = sortByObj || {}, asc = _a.asc, desc = _a.desc;
    var order = asc ? 1 : -1;
    var sortBy = (asc || desc);
    // Validate object config
    throwInvalidConfigErrorIfTrue(!sortBy, 'Expected `asc` or `desc` property');
    throwInvalidConfigErrorIfTrue(asc && desc, 'Ambiguous object with `asc` and `desc` config properties');
    var comparer = sortByObj.comparer && castComparer(sortByObj.comparer);
    return { order: order, sortBy: sortBy, comparer: comparer };
};
// >>> SORTERS <<<
var multiPropertySorterProvider = function (defaultComparer) {
    return function multiPropertySorter(sortBy, sortByArr, depth, order, comparer, a, b) {
        var valA;
        var valB;
        if (typeof sortBy === 'string') {
            valA = a[sortBy];
            valB = b[sortBy];
        }
        else if (typeof sortBy === 'function') {
            valA = sortBy(a);
            valB = sortBy(b);
        }
        else {
            var objectSorterConfig = unpackObjectSorter(sortBy);
            return multiPropertySorter(objectSorterConfig.sortBy, sortByArr, depth, objectSorterConfig.order, objectSorterConfig.comparer || defaultComparer, a, b);
        }
        var equality = comparer(valA, valB, order);
        if ((equality === 0 || (valA == null && valB == null)) &&
            sortByArr.length > depth) {
            return multiPropertySorter(sortByArr[depth], sortByArr, depth + 1, order, comparer, a, b);
        }
        return equality;
    };
};
function getSortStrategy(sortBy, comparer, order) {
    // Flat array sorter
    if (sortBy === undefined || sortBy === true) {
        return function (a, b) { return comparer(a, b, order); };
    }
    // Sort list of objects by single object key
    if (typeof sortBy === 'string') {
        throwInvalidConfigErrorIfTrue(sortBy.includes('.'), 'String syntax not allowed for nested properties.');
        return function (a, b) { return comparer(a[sortBy], b[sortBy], order); };
    }
    // Sort list of objects by single function sorter
    if (typeof sortBy === 'function') {
        return function (a, b) { return comparer(sortBy(a), sortBy(b), order); };
    }
    // Sort by multiple properties
    if (Array.isArray(sortBy)) {
        var multiPropSorter_1 = multiPropertySorterProvider(comparer);
        return function (a, b) { return multiPropSorter_1(sortBy[0], sortBy, 1, order, comparer, a, b); };
    }
    // Unpack object config to get actual sorter strategy
    var objectSorterConfig = unpackObjectSorter(sortBy);
    return getSortStrategy(objectSorterConfig.sortBy, objectSorterConfig.comparer || comparer, objectSorterConfig.order);
}
var sortArray = function (order, ctx, sortBy, comparer) {
    var _a;
    if (!Array.isArray(ctx)) {
        return ctx;
    }
    // Unwrap sortBy if array with only 1 value to get faster sort strategy
    if (Array.isArray(sortBy) && sortBy.length < 2) {
        _a = sortBy, sortBy = _a[0];
    }
    return ctx.sort(getSortStrategy(sortBy, comparer, order));
};
function createNewSortInstance(opts) {
    var comparer = castComparer(opts.comparer);
    return function (arrayToSort) {
        var ctx = Array.isArray(arrayToSort) && !opts.inPlaceSorting
            ? arrayToSort.slice()
            : arrayToSort;
        return {
            asc: function (sortBy) {
                return sortArray(1, ctx, sortBy, comparer);
            },
            desc: function (sortBy) {
                return sortArray(-1, ctx, sortBy, comparer);
            },
            by: function (sortBy) {
                return sortArray(1, ctx, sortBy, comparer);
            },
        };
    };
}
var defaultComparer = function (a, b, order) {
    if (a == null)
        return order;
    if (b == null)
        return -order;
    if (typeof a !== typeof b) {
        return typeof a < typeof b ? -1 : 1;
    }
    if (a < b)
        return -1;
    if (a > b)
        return 1;
    return 0;
};
var sort = createNewSortInstance({
    comparer: defaultComparer,
});
createNewSortInstance({
    comparer: defaultComparer,
    inPlaceSorting: true,
});

function handleValue$1(options) {
  const { addToUsedBy, id, magicString, node, raw, replaceMap, offset = 0 } = options;
  let value = raw;
  const arr = sort(shared.splitCode(value)).desc((x) => x.length);
  for (const str of arr) {
    if (replaceMap.has(str)) {
      addToUsedBy(str, id);
      const v = replaceMap.get(str);
      if (v) {
        value = value.replaceAll(str, v);
      }
    }
  }
  if (typeof node.start === "number" && typeof node.end === "number" && value) {
    magicString.update(node.start + offset, node.end - offset, value);
  }
}
const plugin = helperPluginUtils.declare((api, options) => {
  api.assertVersion(7);
  const { magicString, replaceMap, id, addToUsedBy } = options;
  return {
    visitor: {
      StringLiteral: {
        enter(p) {
          const node = p.node;
          handleValue$1({
            addToUsedBy,
            id,
            magicString,
            node,
            raw: node.value,
            replaceMap,
            offset: 1
          });
        }
      },
      TemplateElement: {
        enter(p) {
          const node = p.node;
          handleValue$1({
            addToUsedBy,
            id,
            magicString,
            node,
            raw: node.value.raw,
            replaceMap,
            offset: 0
          });
        }
      }
    }
  };
});
function preProcessJs(options) {
  const { code, replaceMap, id, addToUsedBy } = options;
  const magicString = new MagicString__default(code);
  babel__default.transformSync(code, {
    presets: [
      // ['@babel/preset-react', {}],
      [
        require("@babel/preset-typescript"),
        {
          allExtensions: true,
          isTSX: true
        }
      ]
    ],
    plugins: [
      [
        plugin,
        {
          magicString,
          replaceMap,
          id,
          addToUsedBy
        }
      ]
    ],
    filename: id
  });
  return magicString.toString();
}

function handleValue(raw, node, options) {
  const { replaceMap, classGenerator: clsGen, splitQuote = true } = options;
  const array = shared.splitCode(raw, {
    splitQuote
  });
  let rawString = raw;
  for (const v of array) {
    if (replaceMap.has(v)) {
      let ignoreFlag = false;
      if (Array.isArray(node.leadingComments)) {
        ignoreFlag = node.leadingComments.findIndex((x) => x.value.includes("tw-mangle") && x.value.includes("ignore")) > -1;
      }
      if (!ignoreFlag) {
        rawString = rawString.replace(shared.makeRegex(v), clsGen.generateClassName(v).name);
      }
    }
  }
  return rawString;
}
function jsHandler(rawSource, options) {
  const result = babel.transformSync(rawSource, {
    babelrc: false,
    ast: true,
    plugins: [
      () => {
        return {
          visitor: {
            StringLiteral: {
              enter(p) {
                const n = p.node;
                n.value = handleValue(n.value, n, options);
              }
            },
            TemplateElement: {
              enter(p) {
                const n = p.node;
                n.value.raw = handleValue(n.value.raw, n, options);
              }
            },
            CallExpression: {
              enter(p) {
                const calleePath = p.get("callee");
                if (calleePath.isIdentifier() && calleePath.node.name === "eval") {
                  p.traverse({
                    StringLiteral: {
                      enter(s) {
                        const res = jsHandler(s.node.value, options);
                        if (res.code) {
                          s.node.value = res.code;
                        }
                      }
                    }
                  });
                }
              }
            }
            // noScope: true
          }
        };
      }
    ],
    minified: options.minified ?? isProd(),
    sourceMaps: false,
    configFile: false
  });
  return result;
}

exports.ClassGenerator = shared.ClassGenerator;
exports.cssHandler = cssHandler;
exports.handleValue = handleValue;
exports.htmlHandler = htmlHandler;
exports.jsHandler = jsHandler;
exports.preProcessJs = preProcessJs;
